/******************
*******************
  creating, deleting, initializing structures
  of points and segments
*******************
******************/

/***********************************************************
Structures allocation and destruction function
***********************************************************/




/***************
 * FUNCTION:     Pt_cree
 * PURPOSE:      creates an array of n points
                 and returns its pointer
 * ARGUMENTS:    long n :
 * ALGORITHM:    
 * RETURN:       p2d *
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

p2d *Pt_cree(long n);


/***************
 * FUNCTION:     Sgl_cree
 * PURPOSE:      creates a segment structure and returns 
                 its pointer. 
		 sets the extremity connections to NULL
 * ARGUMENTS:    None
 * ALGORITHM:    
 * RETURN:       Sgl *
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

Sgl *Sgl_cree(void);


/***************
 * FUNCTION:     Sgl_free
 * PURPOSE:      free segment structure
 * ARGUMENTS:    Sgl *s :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

void Sgl_free(Sgl *s);



/***************
 * FUNCTION:     Imgl_cree
 * PURPOSE:      creates and initialize a structure image laser
                 array of points is set to NULL
		 list of segments is setto NULL
		 coordinate system is set to (0,0,0,0) and 
		 defined as absolute by default (coosys = 'A')
		 iniializes the varaiance of the robot position
		 to 0.
 * ARGUMENTS:    None
 * ALGORITHM:    
 * RETURN:       igla *
 * SIDE EFFECT:  
 * CALLS:        mx_init/matrix.c
 * CALLED BY:    
 ***************/

igla *Imgl_cree(void);
    

/***************
 * FUNCTION:     Imgl_free
 * PURPOSE:      free the image *ig and all of its dynamic components
                 array of points and list of segments
 * ARGUMENTS:    igla *ig :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        dlib_list_el()/dbliste.c
 * CALLED BY:    
 ***************/

void Imgl_free(igla *ig);
  



/**************************************************************
Data Acquisition Functions
   Works with a set of two functions
   -image structure creation and filling;
   -point array creation and filling

Pixel acquisition:

    Imgl_cam_fill
    ReadCamPer

Cartesian coordinates acquisition
    Imgl_cart_fill
    ReadCamCart

**************************************************************/



/***************
 * FUNCTION:     SetMaxPtDist
 * PURPOSE:      set the paramter MaxPtDist (tenth of inch) 
                 all points further than MaxPtDist from the robot center
		 will be discarded during laser data acquisition process
		 in ReadCamCart().
 * ARGUMENTS:    long maxdt :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

void SetMaxPtDist(long maxdt);


/******
Create an array of points and 
fill it with the perception directly from the camera (pixels)
You should have done Set_ls_pix(482) before calling this function
*******/


/***************
 * FUNCTION:     ReadCamPer
 * PURPOSE:      creates an array of points "p2d" and fill it with 
                 the values in the laser value array "laser_val".
		 It is intended only for direct pixel reading.
		 Variance matrices are computed assuming a uniform distribution
		 one pixel wide in "y" coordinate and a gaussian distribution
		 of variance 1 in x coordinate. 
		 x represents the laser value itself
		 y represents the scan line of the camera.
		 Typical use after a gs() call:
		      ReadCamPer(Laser[0],&Laser[1]);
		 The function also computes polar transformations 
		 WARNING: this function assumes you took 482 points and is mainly
		 intended for calibration purposes. 
		 
 * ARGUMENTS:    long nbp :
 *               long *laser_val :
 * ALGORITHM:    
 * RETURN:       p2d *
 * SIDE EFFECT:  
 * CALLS:        Pt_cree() , point_inc_xyro()/las_unc.c
 * CALLED BY:    Imagl_cam_fill()
 ***************/


p2d *ReadCamPer(long nbp,long *laser_val);
  

/***************
 * FUNCTION:     Imgl_cam_fill
 * PURPOSE:      fills the igla structure "*ig" with the
                 the laser data contained in "*laser_val" directly.

		 
		 The coordinate system is set to be absolute: 
		 ig->coosys = 'A'
		 In case you use this with the robot directly, the 
		 functions "Set_pos_image" and "Set_enc_image" allow
		 you to save encoder values in the *igla structure.
		 Here, all values are set to 0 (encoder position
		 + encoder angles + absolute position + absolute angles)

		 ig->x, ig->y = x and y absolute  of the robot
		               at perception time
		 ig->cs,ig->sn = cos and sine of encoder sterring angle
		 ig->cst, ig->snt =  cos and sine of encoder turret angle
		 ig->enc = rob_enc : x,y,st,tu
 
		 
		 All fields ->x,->y,->cs etc are intended to contain absolute
		 coordinates when different from the robots' starting point.
		 
		 ig->enc[4] should always keep memory
		 of the encoder values at the time the perception was made.
		 
 * ARGUMENTS:    igla *ig :
 *               long nbp :
 *               long *laser_val :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        ReadCamPer(), mx_init/matrix.c
 * CALLED BY:    
 ***************/

void Imgl_cam_fill(igla *ig,long nbp,long *laser_val);
    

/***************
 * FUNCTION:     ReadCamCart
 * PURPOSE:      Creates an array of points
                 Fill it with the current perception in physical space
		 (cartesian + conversion to polar )
		 The points are expressed in [position + turret angle]
		 reference frame.
 * ARGUMENTS:    long nbpt :
 *               long *laser_val :
 * ALGORITHM:    the function takes the array "laser_val".
                 Basically,  laser val is of the type Laser[] produced
		 by the robot. 
		 A typical way to use directly with the robot would be:

		 gs();
		 Pt_array = ReadCamCart(Laser[0],Laser + 1);

 * RETURN:       p2d *
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

p2d *ReadCamCart(long nbpt, long *laser_val);





/***************
 * FUNCTION:     Imgl_cart_fill
 * PURPOSE:      fills the igla structure "*ig" with the
                 the laser data contained in "*laser_val".
		 Here, the laser data is transformed into
		 physical coordinates via the calibration 
		 file "laser.cal".
 		 Before a call to this function, you have to call 
		 Read_fic_calib() (see below).

		 Points further than MaxPtDist will be discarded.
		 
		 The coordinate system is set to be the turret: 
		 ig->coosys = 'T'

		 In case you use this with the robot directly, the 
		 functions "Set_pos_image" and "Set_enc_image" allow
		 you to save encoder values in the *igla structure.
		 Here, all values are set to 0 (encoder position
		 + encoder angles + absolute position + absolute angles)

		 ig->x, ig->y = x and y absolute  of the robot
		               at perception time
		 ig->cs,ig->sn = cos and sine of encoder sterring angle
		 ig->cst, ig->snt =  cos and sine of encoder turret angle
		 ig->enc = rob_enc : x,y,st,tu
 
		 
		 All fields ->x,->y,->cs etc are intended to contain absolute
		 coordinates when different from the robots' starting point.
		 ig->enc[4] should always keep memory
		 of the encoder values at the time the perception was made.
		 
		 

 * ARGUMENTS:    igla *ig :
 *               long nbp :
 *               long *laser_val :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        ReadCamCart(), mx_init/matrix.c
 * CALLED BY:    
 ***************/


void Imgl_cart_fill(igla *ig,long nbp,long *laser_val);



/***********************************************************
  information about structures:
***********************************************************/



/***************
 * FUNCTION:     Pt_pos_x
 * PURPOSE:      returns the absolute x coordinate of the point i
                  of the image igl
 * ARGUMENTS:    igla *igl :
 *               long i :
 * ALGORITHM:    Depending on the coordinate system igl->coosys 
                 used by the image igl, this function computes the 
		 projection of igl->tbp[i] through the coordinates 
		 system defined by:
		 case "T" (turret)
		    i->x,ig->y,ig->cst,ig->snt
		 case "S" (steering)
		    i->x,ig->y,ig->cs,ig->sn
		 case "A" (absolute)
		    identity

 * RETURN:       double 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/


double Pt_pos_x(igla *igl,long i);




/***************
 * FUNCTION:     Pt_pos_y
 * PURPOSE:      returns the absolute y coordinate of the point i
                  of the image igl
 * ARGUMENTS:    igla *igl :
 *               long i :
 * ALGORITHM:    Depending on the coordinate system igl->coosys 
                 used by the image igl, this function computes the 
		 projection of igl->tbp[i] through the coordinates 
		 system defined by:
		 case "T" (turret)
		    i->x,ig->y,ig->cst,ig->snt
		 case "S" (steering)
		    i->x,ig->y,ig->cs,ig->sn
		 case "A" (absolute)
		    identity

 * RETURN:       double 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

double Pt_pos_y(igla *igl,long i);





/***************
 * FUNCTION:     Pt_pos_y
 * PURPOSE:      compute the absolute  coordinates of the point i
                  of the image igl. p[0] = x, p[1] = y.
 * ARGUMENTS:    igla *igl :
 *               long i :
                 double *p:
 * ALGORITHM:    Depending on the coordinate system igl->coosys 
                 used by the image igl, this function computes the 
		 projection of igl->tbp[i] through the coordinates 
		 system defined by:
		 case "T" (turret)
		    i->x,ig->y,ig->cst,ig->snt
		 case "S" (steering)
		    i->x,ig->y,ig->cs,ig->sn
		 case "A" (absolute)
		    identity

 * RETURN:       double 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

void Pt_pos(igla *igl,p2d *pt,double *p);


/***************
 * FUNCTION:     Sg_pos
 * PURPOSE:      compute the absolute  coordinates of the segment *sg
                  of the image igl. 
		  p[0] = x of first point (sg->first->x)
		  p[1] = y of first point (sg->first->y)
		  p[2] = x of last point (sg->last->x)
		  p[3] = y of last point (sg->last->y)
		  
 * ARGUMENTS:    igla *igl :
 *               long i :
                 double *p:
 * ALGORITHM:    Depending on the coordinate system igl->coosys 
                 used by the image igl, this function computes the 
		 projection of igl->tbp[i] through the coordinates 
		 system defined by:
		 case "T" (turret)
		    i->x,ig->y,ig->cst,ig->snt
		 case "S" (steering)
		    i->x,ig->y,ig->cs,ig->sn
		 case "A" (absolute)
		    identity

 * RETURN:       double 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/


void Sg_pos(igla *igl,Sgl *sg,double *p);





/***************
 * FUNCTION:     Proj_sg
 * PURPOSE:      computes the projection of sg in the coordinate system defined
                 in igl:

		 case "T" (turret)
		    i->x,ig->y,ig->cst,ig->snt
		 case "S" (steering)
		    i->x,ig->y,ig->cs,ig->sn
		 case "A" (absolute)
		    identity
		 Computes the polar representation of the new segment.

                 returns the projected segment;
		 
                 Warning: creates the segment result, you have to free it.
 * ARGUMENTS:    igla *igl :
 *               Sgl *sg :
 * ALGORITHM:    
 * RETURN:       Sgl *
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

Sgl *Proj_sg(igla *igl,Sgl *sg);



/***************
 * FUNCTION:     Proj_ima_s
 * PURPOSE:      projects segments in *igl in the absolute 
                 coordinate system using *igl coordinate system.
		 Copies the resulting segments in image *ig
		 ig coordinate system is 0,0,0,0.
		 ig->enc = igl->enc

		 ig's segments are allocated in this function
		 ig's structure has to be allocated before
		 calling this function.

		 points in igl->tbp (if any) are not projected
		 variance matrices are merely copied without
		 transformations

		 Returns 1 on success
		 Returns 0 if memory allocation pbs.

 * ARGUMENTS:    igla *igl :
 *               igla *ig :
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        mx_copy/matrix.c, Imgl_free
 * CALLED BY:    
 ***************/

long Proj_ima_s(igla *igl,igla *ig);



/***************
 * FUNCTION:     Set_pos_image
 * PURPOSE:      assigns a new absolute coordinate system to 
                 the image *igl.
		
		 th and tu are in radian
		 this new coordinates system is set in the fields
		 igl->x,igl->y : absolute x and y
		 igl->cs, igl->sn : cos and sin of steering angle
		 igl->cst,igl->snt : cos and sin of turret angle

 * ARGUMENTS:    igla *igl :
 *               double x :
 *               double y :
 *               double th :
 *               double tu :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/


void Set_pos_image(igla *igl,double x,double y,double th,double tu);




/***************
 * FUNCTION:     Set_enc_image
 * PURPOSE:      saves the encoder values in *igl
                 *encoders values are supposed to
		 be longs integers in 10th of inches
		 and 10th or degrees.
		 They are converted in inches and radian
		 igl->enc[0] = x
		 igl->enc[1] = y
		 igl->enc[2] = steering angle
		 igl->enc[3] = turret angle
 * ARGUMENTS:    igla *igl :
 *               long *encoders :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

void Set_enc_image(igla *igl,long *encoders);
  

/***************
 * FUNCTION:     Imgl_create_copy
 * PURPOSE:      returns an indexed copy
                 of *igm.
		 The main purpose is to be able to transform,
		 even truncate, the segments in the resulting image
		 while keeping memory about the original segment in
		 *igm.
		 
 * ARGUMENTS:    igla *igm :
 * ALGORITHM:    all segment in igm->lsgl are copied onto
                 igr->lsgl
		 all segments in igr->lsgl contain the pointer
		 on the corresponding *igm segment in their 
		 ->sfirst field
		 the field igr->usr becomes an array representation
		 of the list igr->lsgl.
		 This is mainly done for aspect computation when we want
		 to know from where are coming the local visible segments.
 * RETURN:       igla *
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

igla *Imgl_create_copy(igla *igm);


/**************************************************************
  Transformations according to the calibration
**************************************************************/



/***************
 * FUNCTION:     read_laser_cal
 * PURPOSE:      reads the file specified in the argument
                 sets the parameters CamParam[8] used for 
		 transforming pixel points into physical 
		 coordinates. 
		 sets the variance on the parameters in
		 VarParam[]
		 WARNING: the file must have been produced by
		 the calibration program.
 * ARGUMENTS:    char *file :
 * ALGORITHM:    
 * RETURN:       long: = 0 if file ok
                       = -1 if pb with the file 
 * SIDE EFFECT:  fills up CamParam[] and VarParam[]
                 changes the static variable "File"
 * CALLS:        Read_init_calib()
 * CALLED BY:    
 ***************/

long read_laser_cal(char *file);


/***************
 * FUNCTION:     Read_fic_calib
 * PURPOSE:      interactively asks for a calibration file
                 name and calls Read_init_calib() as above.
		 WARNING: the file must have been produced by
		 the calibration program.
 * ARGUMENTS:    None
 * ALGORITHM:    
 * RETURN:       long = 0 if file ok
                      = -1 if pb with the file 
 * SIDE EFFECT:  fills up CamParam[] and VarParam[]
                 changes the static variable "File"
 * CALLS:        Read_init_calib()
 * CALLED BY:    
 ***************/

long Read_fic_calib(void);




/***************
 * FUNCTION:     Read_init_calib
 * PURPOSE:      read the file "File";
                 by default File is "laser.cal";
		 If you want to use another file, 
		 call read_laser_cal() (linked
		 version) or Read_fic_calib() (interactive 
		 version)
		 WARNING: the file must have been produced by
		 the calibration program.
 * ARGUMENTS:    None
 * ALGORITHM:    
 * RETURN:       long = 0 if file ok
                      = -1 if pb with the file 
 * SIDE EFFECT:  fills up CamParam[] and VarParam[]
                 changes the static variable "File"
 * CALLS:       
 * CALLED BY:    
 ***************/

long Read_init_calib(void);



/***********************
Transformation function according to the calibration ,
   The resulting points are in the turret frame
   (i.e, : x,y and tu)

   yi = scanline
   xi = laser_val[yi]

   *x and *y are the resulting coordinates
***********************/


/***************
 * FUNCTION:     ProjectPhy
 * PURPOSE:      Transformation function according to the calibration ,
                 The resulting points are in the turret frame
                 (i.e, : x,y and tu)

		 yi = scanline
		 xi = laser_val[yi] in PIXEL

		 *x and *y are the resulting coordinates
		 WARNING: a call to Read_init_calib must have been
		 made for this function to compute correct values.
		 
 * ARGUMENTS:    double xi :
 *               double yi :
 *               double *x :
 *               double *y :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

void ProjectPhy(double xi,double yi,double *x,double *y);




/***************
 * FUNCTION:     pixtorange
 * PURPOSE:      Transformation function according to the calibration ,
                 The resulting points are in the turret frame
                 (i.e, : x,y and tu)

		 yi = scanline
		 xi = laser_val[yi] in PIXEL

		 *x and *y are the resulting coordinates
		 Same as above but xi and yi are integers and 
		 x and y represents tenth of the basic unit of
		 the calibration.

		 WARNING: a call to Read_init_calib must have been
		 made for this function to compute correct values.
		 
 * ARGUMENTS:    long xi :
 *               long yi :
 *               long *x :
 *               long *y :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/


void pixtorange(long xi,long yi,long *x,long *y);




/*************
  same thing as ProjectPhy but computes also
  the variance on x and y in vx,vxy and vy
*************/


/***************
 * FUNCTION:     ProjectPhyP
 * PURPOSE:      Transformation function according to the calibration 
                 Physical points + Variance.

                 The resulting points are in the turret frame
                 (i.e, : x,y and tu)

		 yi = scanline
		 xi = laser_val[yi] in PIXEL

		 *x and *y are the resulting coordinates
		 vx,vxy and vy represents the covariance
		 matrix :

		 [vx vxy]
		 [vxy vy]

		 WARNING: a call to Read_init_calib must have been
		 made previously for this function to compute correct values.
		 
 * ARGUMENTS:    double xi :
 *               double yi :
 *               double *x :
 *               double *y :
 *               double *vx :
 *               double *vxy :
 *               double *vy :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:   mxfs_mmulttr()/matrix.c     
 * CALLED BY:    
 ***************/


void ProjectPhyP(double xi,double yi,double *x,double *y,double *vx,
		    double *vxy,double *vy);

















