/*
 *         Laser Data Segmentation
 *
 * two versions:
 *  segmentation(igl,option)
 * 
 *      input: an image of type "*igla"
 *          This image contain an array of igl->nbp
 *          raw points :  igl->tbp
 *      output: the same image with the list igl->lsgl of
 *           segment computed from the points igl->tbp.
 *
 *      option: 0 for least square estimation
 *           1 for kalman estimation
 *   
 *  get_sg_host(sgt,nbp,laser_val)
 *       
 *      input: the number of point "nbp" in the 
 *             laser pixel array "laser_val"
 *        
 *      output: the array of segment "sgt"
 *              and the number of segment (returned value)
 *      
 *
 *    Points are assumed coming from the same central position and
 *    ordered in angle in a trigonometric manner.
 *    
 *
 *   function conf_sg_host:
 *                         sets the parameters:
 *
 *                           thres (on the least square error)
 *                           nbmin (minimum number of points in
 *                                  a segment)
 *                           gap (maximum space between two consecutive
 *                                points in a segment)
 *
 *   function segmentation: computes the longest segments from
 *                          igl->tbp verifying the constraints set
 *                          by Set_segm_param
 *

 *   function first_discont_point       : returns TRUE if there is a discontinuity
 *                           between the  points j and k. If TRUE,
 *                           m is the discontinuous point.
 *   function first_continuous_point      : returns the index of the next
 *                              point, after k, which could be the
 *                              first point of a segment.
 *
 *   functions EndSS, eEndBP, EndBB :
 *                           computes the different kind of
 *                           intersections
 */



/***************
 * FUNCTION:     get_sg_host
 * PURPOSE:      segmentation using directly the server format
                 of laser values and returning a server format 
		 array of segments.

 * ARGUMENTS:    long *sgt : server format array of segment (output)
 *               long nbp : number of pixel points in laser_val (input)
 *               long *laser_val : server format array of pixels (input)

 * ALGORITHM:    1) transforms the "nbp" pixel points 
		 of laser_val into cartesian
		 data using the current 
		 selected calibration parameters
		 and fills the structure "ig"
		 2) calls segmentation()
		 3) copies the internal structure "ig" into
		 the array "sgt"
		 4) frees the internal structure.
		 if conf_sg_host  has been called like this:
		 conf_sg_host(a,nb,c), sgt must be an array 
		 of (4*nbp)/(nb-1) + 2 integers
		 
		 The function returns the number of segments.
		 It uses the function "segmentation" above.
		 
		 The number of segments will also be in sgt[0].
		 
		 Note: this segmentation will use only least square.

 * RETURN:       long : number of segments in "sgt"
 * SIDE EFFECT:  
 * CALLS: Imgl_cree()/las_struct.c, Imgl_free()/las_struc.c,
          Imgl_cart_fill()/las_struct.c, segmentation()
 * CALLED BY:    
 ***************/


static long get_sg_host(long *sgt,long nbp, long *laser_val );



/***************
 * FUNCTION:     first_continuous_point
 * PURPOSE:      returns the index of the next
                 point, after k, which could be the
		 first point of a segment.
		 This function looks for a point such that 
		 the following point is continuous to it, in the same
		 sens than discont right above.
 * ARGUMENTS:    int k :
 * ALGORITHM:    
 * RETURN:       int 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:  segmentation()  
 ***************/

static int first_continuous_point(int k);


/***************
 * FUNCTION:     alined
 * PURPOSE:      determine weither the points from first to
                 last are alined (mean square error < thres).
		 Points first to j are already integrated in
		 the inertia matrix *mic.
		 The line parameters are computed in *dt (see lasdef.h).
		 Returns 1 if alined, 0 if not.
		 In case points are alined, updates the inertia matrix
		 "*mic".
		 NOTE: the inertia matrix doesn't have anything to do 
		 with the matrix representation in matrix.c 

 * ARGUMENTS:    int first : index of the first point integrated in *mic
 *               int last : index of the last point already integrated
 *               int attempt : index of the point
 *               struct matiner *mic : accumulation of first and second
                                       moments of inertia of points
				       belonging to the current segment
 *               Sgl *dt : current infinite line
 * ALGORITHM:    
 * RETURN:       int 
 * SIDE EFFECT:  
 * CALLS: lsqr_approx()        
 * CALLED BY: segmentation()   
 ***************/


static int alined (int first,int last,int attempt,struct matiner *mic,
	Sgl *dt);



/***************
 * FUNCTION:     stepping_back
 * PURPOSE:      looks for the furthest point between pointsloc[last] and
                 pointsloc[attemp-1] from the line joining pointsloc[first]
		 to pointsloc[attempt].
		 This function is called only when the goup of points 
		 first->attempt has been determined as non-alined.
		 The furthest point returned  will then be taken as a new 
		 possible limit for the current segment.

 * ARGUMENTS:    int first :
 *               int last :
 *               int attempt :
 * ALGORITHM:    
 * RETURN:       int 
 * SIDE EFFECT:  
 * CALLS:  dt2pt/lasutil.c      
 * CALLED BY: segmentation()   
 ***************/


static int stepping_back(int first,int last,int attempt);





/***************
 * FUNCTION:     first_discont_point
 * PURPOSE:      looks if there is a discontinuity
                 between pointsloc[j] and pointsloc[k]
		 If yes, put the discontinuous point 
		 index in *m.
		 (means pointsloc[*m] cannot belong 
		 to the same segment as any points 
		 pointsloc[n], j <= n < *m)
		 Returns 1 if a discontunuity exists,
		         0 if not
		 A discontinuity between points i and n
		 occurs if:

		 - there is a local specularity angle of 
		   less than 0.5 degrees and the local difference in 
		   distance between i and n is more than D_MIN
		 or if:

		 - the distance between i and n is more than the gap
		   defined bye the user in conf_sg

		 or if:
		 
		 - the point n is at a distance less than SATU or
		   more than DISMAX, meaning this point can't be 
		   valid
		
 * ARGUMENTS:    int j :
 *               int k :
 *               int *m :
 * ALGORITHM:    
 * RETURN:       int 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY: segmentation()   
 ***************/

static int first_discont_point (int j,int k,int *m);




/***************
 * FUNCTION:     first_continuous_point
 * PURPOSE:      returns the index of the next
                 point, after k, which could be the
		 first point of a segment.
		 This function looks for a point such that 
		 the following point is continuous to it, in the same
		 sens than first_discont_point right above.
 * ARGUMENTS:    int k :
 * ALGORITHM:    
 * RETURN:       int 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:  segmentation()  
 ***************/

int first_continuous_point(int k);



/***************
 * FUNCTION:     extremities
 * PURPOSE:      least mean square estimation and kalman
                 compute only the supporting line of the segments
		 This function computes the extremities.
              
 * ARGUMENTS:    igla *ig :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:   EndBlast(),EndBfirst(),EndVV_BB()
 * CALLED BY: segmentation()        
 ***************/

static void extremities(igla *ig);


/*****
  uses function of lasutil.c for intersecting segments
*****/


/***************
 * FUNCTION:     EndVV_BB
 * PURPOSE:      tests if the intersection between psglp
                 and psglc is "reasonably" a vertex. 
		 By reasonable, we mean:
		 - the end of psglp and the begining of psglc
		   are close (angle < 3.5 degrees and 
		   distance < 3 inches)
		 - the intersection point is close to the mean 
		 angle between the two extremities (no more than 5.5 degrees)

		 If the conditions are fulfilled, 
		      psglp->last and psglc->first are set to be the same
		      psglp->slast points to psglc
		      psglc->sfirst points to psglp
		      psglp->typl = psglc->typf = 'V'

 * ARGUMENTS:    Sgl *psglp :
 *               Sgl *psglc :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        EndBB(), intersg()
 * CALLED BY:    extremities()
 ***************/


static void EndVV_BB (Sgl *psglp, Sgl *psglc);



/***************
 * FUNCTION:     EndBB
 * PURPOSE:      computes the border points of
                 the end of psglp and the begining 
		 of psglc
 * ARGUMENTS:    Sgl *psglp :
 *               Sgl *psglc :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        EndBlast(), EndBfirst()
 * CALLED BY:    EndVV_BB()
 ***************/

static void EndBB (Sgl *psglp,Sgl *psglc);
 
 

static void EndBfirst (Sgl *sg);
     

static void EndBlast (Sgl *sg);

   

static void mazmi(struct matiner *mi);




/**************
  computes a least-square estimation of the points
  from tab. 
  n is the number of points already
  integrated;

  mi is the matrix of inertia of those points

  p is the number of points to integrate

  tab is the array of points to be integrated

  dte is the line we want to be computed

  err is the quadratic error;

  Side effects:

  dte (->vx,->vy and ->d, ->er) is updated such that d > 0
  and  || vx^2 + vy^2 || = 1;

  mi is updated;

***************/
  

static void lsqr_approx(int *n,struct matiner *mi,int p,p2d *tab,Sgl *dte,double *err);





























